/* fortuned is a simple, educational daemon, written as accompanying code     */
/* to my 'daemons and other monsters' for the haifa linux club                */
/* (see http://linuxclub.il.eu.org or htp://www.pointer.co.il/~mulix/)        */
/*                                                                            */
/* it creates a fortune device, and thereafter, any process which reads from  */
/* this device will get a fortune. read the code for further details.         */
/*                                                                            */
/* thanks to raxy and sectorx for helpful feedback.                           */
/* thanks to guy (choo@actcom.co.il) for feedback and guidance                */
/* mulix <mulix@actcom.co.il                                                  */ 

#include <stdio.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <signal.h>
#include <fcntl.h>

#include "zmalloc.h"
#include "util.h"

#define MAXFD 64
#define DEFAULT_FORTUNE_DEVICE "/dev/fortune"

volatile int dead_pipe = 0;

const char* 
get_fortune_device()
{
     static char* dev_file = NULL;
     char* dev_file_env = NULL;

     /* if we've already built the device file, return it */
     if (dev_file)
	  return dev_file;

     /* has the user set a FORTUNE_DEVICE environment variable for us? */
     dev_file_env = getenv("FORTUNE_DEVICE");

     /* guess not */
     if (!dev_file_env){
	  /* make enough room in the buffer for the DEFUALT_FORTUNE_DEVICE */
	  dev_file = zmalloc(strlen(DEFAULT_FORTUNE_DEVICE)+1);
     }else{
	  /* yup. make enough room in the buffer for the user's environment var */
	  dev_file = zmalloc(strlen(dev_file_env)+1);
     }

     /* make our buffer contain the correct path to the device */
     sprintf(dev_file, "%s", 
	     dev_file_env ? dev_file_env : DEFAULT_FORTUNE_DEVICE);

     /* and return it */
     return dev_file;
}

static void 
cleanup(const char* err)
{
     unlink(get_fortune_device());
     fatal((char*)err);
}

static void 
sighandler(int signum)
{
     char errstring[256];
     snprintf(errstring, sizeof(errstring), "caught signal '%d', dying", signum);
     cleanup(errstring);
}

static void
check_fortune_device(const char* fortune_dev)
{
     struct stat buf;
     int rc = -1;

     assert(fortune_dev);
     
     log("checking for existence of the device file '%s'", fortune_dev);

     /* stat our file, to check if it exists */
     rc = stat(fortune_dev, &buf);

     /* excellent, it doesn't exit. return from the function */
     if (rc == -1)
	  return;
     
     /* yes, it exists. not good, we bail out */
     fatal("the fortune device file '%s' exists?!\n"
	   "please remove it or specify a different FORTUNE_DEVICE environment variable. ", fortune_dev);
}

/* the functio used to turn us into a daemon */
static void 
daemonize()
{
     int i;
     pid_t pid;

     /* fork for the first time. */ 
     if ((pid = fork()) != 0){
	  if (pid == -1){
	       warn("failed to fork into background! error '%s'", strerror(errno));
	       warn("reaining in foreground");
	       return;
	  }
	  /* parent, exit */
	  exit(0);
     }

     /* we're now in the first child */
     /* setsid() will create a new session for us, since 
	because of the earlier fork() we're not a process group leader.
	this will assure that we have no controlling terminal */
     setsid();

     /* ignore SIGHUP signals */
     signal(SIGHUP, SIG_IGN);

     /* fork again */
     /* this guarantees that we will never have a controlling terminal, */
     /*	since we want be the session group leader anymore               */
     if ((pid = fork()) != 0){
	  if (pid == -1){
	       warn("failed to fork into background! error '%s'", strerror(errno));
	       warn("reaining in foreground");
	       return;
	  }
	  /* parent, exit */
	  exit(0);
     }

     /* change our working directory to /, to make sure
	umounting the file system where we started is possible */
     chdir("/");
     
     /* set out umask to 0, to make sure we fully control our
	file creating bits */
     umask(0);

     /* close all (possibly) open file desriptors */
     /* TODO: calling getdtablesize() here loops endlessly. why? */
     for (i = 0; i < MAXFD ; i++)
	  close(i);
     
     log("succesfully daemonized");
}

static void 
sigpipehandler(int signum)
{
     assert(signum == SIGPIPE);
     dead_pipe = 1;
}

static void 
handle_commands(const char* device_file)
{
     int rc;
     int fd = -1;
     char errstring[256];
     int counter = 0;

#define OPEN_FLAGS      (O_WRONLY)
#define OPEN_MODE       (O_NOCTTY | O_CREAT)
#define OPEN_ATTRIBUTES (S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
#define OPEN_COMMON     (OPEN_MODE | OPEN_ATTRIBUTES)
   
     if (!device_file)
	  fatal("handle_commands called with an invalid device_file "
		"parameter... can't go on like this. goodbye.");
     
     /* open the FIFO */
     log("making the FIFO at '%s'", device_file);
     rc = mkfifo(device_file, OPEN_COMMON);
     if (rc == -1)
	  fatal("failed to mkfifo. error '%s'", strerror(errno));

     /* set our behaviour in regards to signals */
     signal(SIGINT, sighandler);
     signal(SIGKILL, sighandler);
     signal(SIGTERM, sighandler);
	    
     /* this is the main loop */
     while(1){
	  FILE* pipe_in;
	  char readbuf[1024];
	  fd_set wset;
	  /* first of all, open the device file for writing */
	  fd = open(device_file, OPEN_FLAGS, OPEN_COMMON);
	  log ("succesfully opened!");
	  if (fd == -1){
	       snprintf(errstring, sizeof(errstring), 
			"failed to open the fifo. error '%s'", strerror(errno));
	       goto cleanup;
	  }

/* when using a pipe, we don't need to select, as open will only return once we have 
   a reader for our pipe. therefor, as soon as open returns, we can go on and no 
   need to wait for the pipe. if we were using a socket however, the select would 
   definitely have been needed. */
#define USING_PIPE 0	  
#ifdef USING_PIPE
	  FD_ZERO(&wset);
	  FD_SET(fd, &wset);
	  /* now wait until someone tries to read from it */
	  rc = select(fd + 1, NULL, &wset, NULL, NULL);
	  /* ooh, a nasty error */
	  if (rc == -1){
	       /* ignore interruption by non caught signals */
	       if (errno == EINTR){
		    continue;
	       }
	       snprintf(errstring, sizeof(errstring), 
			"failed to select. error '%s'", strerror(errno));
	       goto cleanup;
	  }
	  /* a timeout?! didn't we specifiy no timeouts? */
	  if (rc == 0){
	       snprintf(errstring, sizeof(errstring), 
			"select returned timeout?!");
	       goto cleanup;
	  }

	  /* select returned, but our file desciptor isn't set? very peculiar */
	  if (!FD_ISSET(fd, &wset)){
	       snprintf(errstring, sizeof(errstring), 
			"select returned but fd is not set?!");
	       goto cleanup;
	  }
#endif

	  /* finally, let's do our stuff */
	  /* first of all, make sure we handle SIGPIPE gracefully */
	  signal(SIGPIPE, sigpipehandler);
	  
	  /* we open a pipe with the fortune binary */
	  if ((pipe_in = popen(FORTUNE_BINARY, "r")) == NULL){
	       snprintf(errstring, sizeof(errstring), 
			"could not popen(%s), error '%s'", FORTUNE_BINARY,strerror(errno));
	       goto cleanup;
	  }
	  
	  /* and while it has stuff to write */
	  while (!dead_pipe){
	       int pfd = fileno(pipe_in);
	       int wrc = 0;
	       /* we read from it */
	       memset (readbuf, 0, sizeof (readbuf));
	       rc = read(pfd, readbuf, sizeof(readbuf));
	       if (rc == -1){
		    if (errno == EINTR)
			 continue;
		    snprintf(errstring, sizeof(errstring), 
			     "read returned error '%s'", strerror(errno));
		    goto cleanup;
	       }else if (rc == 0){
		    /* EOF is reached, hooray. */
		    break;
	       }
	       /* and write it to our caller */
	       while (wrc < rc){
		    int bytes_written = 0; 
		    if ((bytes_written = write (fd, readbuf+wrc, strlen(readbuf)-wrc)) == -1){
			 if ((errno == EINTR) || (errno == EAGAIN))
			      continue;
			 warn("failed to write to the fd, error '%s'", strerror(errno));
			 goto pipe_done;
		    }
		    wrc += bytes_written;
	       }
	  }
     pipe_done:
	  /* let our caller know that this is it */
	  close(fd);
	  usleep(10);
	  /* we don't need the pipe to the fortune binary anymore */
	  pclose(pipe_in);

	  /* restore SIGPIPE handling to default */
	  signal(SIGPIPE, SIG_DFL);

	  /* tada! let's brag in the logs */
	  log("fortuned: %d satisfied customers serverd!", ++counter);
     }
 cleanup:
     if (fd)
	  close(fd);
     cleanup(errstring);
}

static void 
parse_command_line(int argc, char** argv, int* foreground, int* sleep)
{
     while (--argc){
	  argv++;
	  if ((*argv)[0] == '-'){
	       if ((*argv)[1] == 'f'){
		    *foreground = 1;
		    continue;
	       }
	       if ((*argv)[1] == 's'){
		    *sleep = 1;
		    continue;
	       }
	  }
     }
}

static void 
check_fortune_binary()
{
     struct stat buf;
     int rc = -1;

     assert(FORTUNE_BINARY);
     
     log("checking for existence of the the fortune binary '%s'", FORTUNE_BINARY);

     /* stat our file, to check if it exists */
     rc = stat(FORTUNE_BINARY, &buf);

     if (rc == -1)
	  /* well, it doesn't exist or is not readable. not good, we bail out */
	  fatal("the fortune binary '%s' does not exists!\n"
		"make sure to specify the correct location in the Makefile ", FORTUNE_BINARY);

     /* check for executablity */
     if (!(buf.st_mode & S_IXUSR))
	  fatal("the fortune binary '%s' is not executable!\n" 
		"make sure to specify the correct executable in the Makefile ", FORTUNE_BINARY);	  
}

int 
main(int argc, char** argv)
{
     int stay_in_foreground = 0;
     int sleep_on_start = 0;
     const char* device_file;
     parse_command_line(argc, argv, &stay_in_foreground, &sleep_on_start);

     open_log(argv[0]);

     device_file = get_fortune_device();
     check_fortune_device(device_file);
     check_fortune_binary();

     if (!stay_in_foreground){
	  daemonize();
	  if (sleep_on_start){
	       log("daemonized, now pid is '%d'", getpid());
	       sleep(20);
	  }
     }

     handle_commands(device_file);
     close_log();
     return 0;
}
